
  /////////////////////////////////////////////////////
  /////////////////////////////////////////////////////
  ////                                             ////
  ////        HDR Alignment Tool Sourcefile        ////
  ////         Free for non-commercial use         ////
  ////                                             ////
  ////        Copyright  2007 Boyd Eagleson       ////
  ////    (Contact... 'Eggy Boil' on Flickr.com)   ////
  ////                                             ////
  /////////////////////////////////////////////////////
  /////////////////////////////////////////////////////

unit ImageRoutines;

interface

uses
  Graphics, Jpeg, SysUtils, InterfaceUnit, Math, Forms, Controls, Dialogs;

procedure ImageToSpherical(ImageX   : Integer;
                           ImageY   : Integer;
                       var Azimuth  : Extended;
                       var Meridian : Extended);

procedure SphericalToImage(Azimuth  : Extended;
                           Meridian : Extended;
                       var ImageX   : Integer;
                       var ImageY   : Integer);

procedure RealImageToSpherical(ImageX   : Extended;
                               ImageY   : Extended;
                           var Azimuth  : Extended;
                           var Meridian : Extended);

procedure SphericalToRealImage(Azimuth  : Extended;
                               Meridian : Extended;
                           var ImageX   : Extended;
                           var ImageY   : Extended);

function SphericalLength(Side_a  : Extended;
                         Side_b  : Extended;
                         Angle_C : Extended): Extended;

function SphericalAngleControlPoint(CP2_Azim   : Extended;
                                    CP_Length  : Extended;
                                    CP1_Azim   : Extended;
                                    PoleAngle  : Extended): Extended;

function SphericalAngleImageStage_1(Point_Azim   : Extended;
                                    Point_Length : Extended;
                                    CP_Azim      : Extended;
                                    PoleAngle    : Extended): Extended;

function SphericalAngleImageStage_2(Point_Length : Extended;
                                    Point_Azim   : Extended;
                                    CP_Azim      : Extended;
                                    TempAngle    : Extended): Extended;

function SphericalAngleImagePoint(Side_a  : Extended;
                        Side_b  : Extended;
                        Side_c  : Extended;
                        Angle_B : Extended): Extended;

procedure FindGreyPoint(Image : TBitmap;
                var GreyPoint : Integer;
                var Histogram : THistogram);

procedure LoadImage(ImageFilename  : String;
                    var InputImage : TBitmap);

procedure DrawZoomedImage(Image : TInputImage;
                var ZoomedImage : TBitmap);

procedure DrawZoomedHistograms(Image : TInputImage;
                    var OutputBitmap : TBitmap);

procedure DrawAlignedPreview(Image1 : TInputImage;
                             Image2 : TInputImage;
                             Image3 : TInputImage;
                             Master : Integer;
                   var OutputBitmap : TBitmap);

procedure BuildAlignedBitmap(Source   : TInputImage;
                             IsMaster : Boolean;
                         var Result   : TBitmap);

procedure DrawQuickPreview(DeltaX        : Integer;
                           DeltaY       : Integer;
                           InputPreview : TBitmap;
                       var Result       : TBitmap);

function Interpolate(A_PixelValue  : Byte;
                     B_PixelValue  : Byte;
                     C_PixelValue  : Byte;
                     D_PixelValue  : Byte;
                     X             : Extended;
                     Y             : Extended): Extended;

function CapPixelValue(Value : Extended): Word;

implementation

uses SaveUnit;


function Luminance(r,g,b: Byte): Integer;
begin
  Result := (Min(r, Min(g, b)) + Max(r, Max(g, b))) div 2;
end;



procedure ImageToSpherical(ImageX   : Integer;
                           ImageY   : Integer;
                       var Azimuth  : Extended;
                       var Meridian : Extended);
var
Sphere_X      : Extended;
Sphere_Y      : Extended;
Sphere_Z      : Extended;
R             : Extended;         // Distance from sphere centre to image pixel
begin
  R := sqrt((ImageX * ImageX) + (ImageY * ImageY) + (Form1.MasterImageDistance * Form1.MasterImageDistance));
  Sphere_X := Form1.MasterImageDistance / R ;
  Sphere_Y := -ImageX / R ;
  Sphere_Z := -ImageY / R ;
  if Sphere_X = 0
    then Meridian := 0
    else Meridian := ArcTan2(Sphere_Y, Sphere_X);
  if Sphere_Z = 0
    then Azimuth := pi/2
    else Azimuth := ArcTan2(sqrt((Sphere_X * Sphere_X) + (Sphere_Y * Sphere_Y)), Sphere_Z);
end;



procedure SphericalToImage(Azimuth  : Extended;
                           Meridian : Extended;
                       var ImageX   : Integer;
                       var ImageY   : Integer);
var
Sphere_X      : Extended;
Sphere_Y      : Extended;
Sphere_Z      : Extended;
R             : Extended;         // Distance from sphere centre to image pixel
begin
  Sphere_X      := Sin(Azimuth) * Cos(Meridian);
  Sphere_Y      := Sin(Azimuth) * Sin(Meridian);
  Sphere_Z      := Cos(Azimuth);
  R             := Form1.MasterImageDistance / Sphere_X;
  // Range Check
  if R >  MaxInt then R :=  MaxInt;
  if R < -MaxInt then R := -MaxInt;
  ImageX        := round(-Sphere_Y * R);
  ImageY        := round(-Sphere_Z * R);
end;


procedure RealImageToSpherical(ImageX   : Extended;
                               ImageY   : Extended;
                           var Azimuth  : Extended;
                           var Meridian : Extended);
var
Sphere_X      : Extended;
Sphere_Y      : Extended;
Sphere_Z      : Extended;
R             : Extended;         // Distance from sphere centre to image pixel
begin
  R := sqrt((ImageX * ImageX) + (ImageY * ImageY) + (Form1.MasterImageDistance * Form1.MasterImageDistance));
  Sphere_X := Form1.MasterImageDistance / R ;
  Sphere_Y := -ImageX / R ;
  Sphere_Z := -ImageY / R ;
  if Sphere_X = 0
    then Meridian := 0
    else Meridian := ArcTan2(Sphere_Y, Sphere_X);
  if Sphere_Z = 0
    then Azimuth := pi/2
    else Azimuth := ArcTan2(sqrt((Sphere_X * Sphere_X) + (Sphere_Y * Sphere_Y)), Sphere_Z);
end;



procedure SphericalToRealImage(Azimuth  : Extended;
                               Meridian : Extended;
                           var ImageX   : Extended;
                           var ImageY   : Extended);
var
Sphere_X    : Extended;
Sphere_Y    : Extended;
Sphere_Z    : Extended;
R           : Extended;         // Distance from sphere centre to image pixel
begin
  Sphere_X  := Sin(Azimuth) * Cos(Meridian);
  Sphere_Y  := Sin(Azimuth) * Sin(Meridian);
  Sphere_Z  := Cos(Azimuth);
  R         := Form1.MasterImageDistance / Sphere_X;
  ImageX    := -Sphere_Y * R;
  ImageY    := -Sphere_Z * R;
end;


function SphericalLength(Side_a  : Extended;
                         Side_b  : Extended;
                         Angle_C : Extended): Extended;
var
Side_c : Extended;
Sin_a  : Extended;
Cos_a  : Extended;
Sin_b  : Extended;
Cos_b  : Extended;
begin
  // Initialise
  SinCos(Side_a, Sin_a, Cos_a);
  SinCos(Side_b, Sin_b, Cos_b);
  Side_c := (Cos_a * Cos_b) + (Sin_a * Sin_b * Cos(Angle_C));
  // Test
  Result := 0;
  if Side_c >  1 then Exit;
  Result := pi;
  if Side_c < -1 then Exit;
  // Calculate length using cosine rule for spherical triangles
  Side_c := ArcCos(Side_c);
  // Assign result
  Result := Side_c;
end;


function SphericalAngleControlPoint(CP2_Azim   : Extended;
                                    CP_Length  : Extended;
                                    CP1_Azim   : Extended;
                                    PoleAngle  : Extended): Extended;
var
CP_Angle : Extended;
SphHyp  : Extended;
begin
  // initialise
  if PoleAngle < 0
    then Result := 3*pi/2
    else Result := pi/2;
  // Test Input Values
  if Sin(CP_Length) = 0 then exit;
  if Abs((Sin(CP2_Azim) * Sin(PoleAngle))) > Abs(Sin(CP_Length)) then exit;
  // Calculate angle using sine rule for spherical triangles
  CP_Angle := ArcSin((Sin(CP2_Azim) * Sin(PoleAngle)) / Sin(CP_Length));
  // Find the azimuth of the point when CP_Angle = 90 (ie Spherical Hypotenuse)
  if CP1_Azim < pi/2
    then SphHyp := ArcTan2(Tan(CP1_Azim), Cos(PoleAngle))
    else SphHyp := ArcTan2(Tan(CP1_Azim), Cos(PoleAngle)) + pi;
  // Test for quadrant and adjust output accordingly
  if CP2_Azim > SphHyp then CP_Angle := pi - CP_Angle;
  if (CP2_Azim < SphHyp) and (PoleAngle < 0) then CP_Angle := (2 * pi) + CP_Angle;
  // Assign result
  Result := CP_Angle;
end;


function SphericalAngleImageStage_1(Point_Azim   : Extended;
                                    Point_Length : Extended;
                                    CP_Azim      : Extended;
                                    PoleAngle    : Extended): Extended;
var
Angle_A : Extended;
SphHyp  : Extended;
begin
  // initialise
  if PoleAngle < 0
    then Result := -pi/2
    else Result := pi/2;
  // Test Input Values
  if Sin(Point_Length) = 0
  then begin
    Result := 0;
    exit;
  end;
  if Abs((Sin(Point_Azim) * Sin(PoleAngle))) > Abs(Sin(Point_Length))
  then begin
    exit;
  end;
  // Calculate angle using sine rule for spherical triangles
  Angle_A := ArcSin((Sin(Point_Azim) * Sin(PoleAngle)) / Sin(Point_Length));
  // Find the azimuth of the point when Angle_A = 90 (ie Spherical Hypotenuse)
  if CP_Azim < pi/2
    then SphHyp := ArcTan2(Tan(CP_Azim), Cos(PoleAngle))
    else SphHyp := ArcTan2(Tan(CP_Azim), Cos(PoleAngle)) + pi;
  // Test for quadrant and adjust output accordingly
  if (Point_Azim > SphHyp) and (PoleAngle > 0) then Angle_A := pi - Angle_A;
  if (Point_Azim > SphHyp) and (PoleAngle <= 0) then Angle_A := -pi - Angle_A;
  // Assign result
  Result := Angle_A;
end;


function SphericalAngleImageStage_2(Point_Length : Extended;
                                    Point_Azim   : Extended;
                                    CP_Azim      : Extended;
                                    TempAngle    : Extended): Extended;
var
Angle_A : Extended;
SphHyp  : Extended;
begin
  // initialise
  if TempAngle < 0
    then Result := -pi/2
    else Result := pi/2;
  // Test Input Values
  if Sin(Point_Length) = 0
  then begin
    Result := 0;
    exit;
  end;
  if Abs((Sin(Point_Length) * Sin(TempAngle))) > Abs(Sin(Point_Azim)) then exit;
  // Calculate angle using sine rule for spherical triangles
  Angle_A := ArcSin((Sin(Point_Length) * Sin(TempAngle)) / Sin(Point_Azim));
  // Find the azimuth of the point when Angle_A = 90 (ie Spherical Hypotenuse)
  if CP_Azim < pi/2
    then SphHyp := ArcTan2(Tan(CP_Azim), Cos(TempAngle))
    else SphHyp := ArcTan2(Tan(CP_Azim), Cos(TempAngle)) + pi;
  // Test for quadrant and adjust output accordingly
  if (Point_Length > SphHyp) and (TempAngle > 0) then Angle_A := pi - Angle_A;
  if (Point_Length > SphHyp) and (TempAngle <= 0) then Angle_A := -pi - Angle_A;
  // Assign result
  Result := Angle_A;
end;


function SphericalAngleImagePoint(Side_a  : Extended;
                        Side_b  : Extended;
                        Side_c  : Extended;
                        Angle_B : Extended): Extended;
var
Angle_A : Extended;
SphHyp  : Extended;
begin
  // initialise
  if Angle_B < 0
    then Result := -pi/2
    else Result := pi/2;
  // Test Input Values
  if Sin(Side_b) = 0
  then begin
    exit;
  end;
  if Abs((Sin(Side_a) * Sin(Angle_B))) > Abs(Sin(Side_b))
  then begin
    exit;
  end;
  // Calculate angle using sine rule for spherical triangles
  Angle_A := ArcSin((Sin(Side_a) * Sin(Angle_B)) / Sin(Side_b));
  // Find the azimuth of the point when Angle_A = 90 (ie Spherical Hypotenuse)
  if Side_c < pi/2
    then SphHyp := ArcTan2(Tan(Side_a), Cos(Angle_B))
    else SphHyp := ArcTan2(Tan(Side_a), Cos(Angle_B)) + pi;
  // Test for quadrant and adjust output accordingly
  if (Side_a > SphHyp) and (Angle_B > 0) then Angle_A := pi - Angle_A;
  if (Side_a > SphHyp) and (Angle_B <= 0) then Angle_A := -pi - Angle_A;
  // Assign result
  Result := Angle_A;
end;


function FindContrastAdjust(GreyAdjust : Integer): Extended;
begin
  if GreyAdjust < 0
  then Result := Power(10, (GreyAdjust + 255) / 255) / 10
  else Result := Power(10, (GreyAdjust + 255) / 255) / 10;
end;


procedure FindGreyPoint(Image : TBitmap;
                var GreyPoint : Integer;
                var Histogram : THistogram);
var
LineIndex   : Integer;
PixelIndex  : Integer;
SourceLine  : PByteArray;    // For Scanline
GreyCount   : Int64;
PixelCount  : Integer;
begin
  GreyCount  := 0;
  PixelCount := 0;
  for PixelIndex := 0 to 255
    do Histogram[PixelIndex] := 0;
  for LineIndex := 0 to Image.Height-1 do
  begin
    SourceLine := Image.ScanLine[LineIndex];
    for PixelIndex := 0 to Image.Width -1 do
    begin
      GreyPoint := Luminance(SourceLine[PixelIndex*3],
                             SourceLine[(PixelIndex*3)+1],
                             SourceLine[(PixelIndex*3)+2]);
      GreyCount := GreyCount + GreyPoint;
      Inc(PixelCount);
      Inc(Histogram[GreyPoint]);
    end;   // for PixelIndex
  end;     // for LineIndex
  if PixelCount = 0
    then GreyPoint := 0
    else GreyPoint := GreyCount div PixelCount;
end;


procedure LoadImage(ImageFilename  : String;
                    var InputImage : TBitmap);
var
JPEGLoad : TJPEGImage;
begin
  try  // Start of EInvalidGraphic Exception handler
    // Clear Input Bitmap
    InputImage.Width := 0;
    InputImage.Height := 0;
    // Check for and load a Bitmap
    if (ExtractFileExt(ImageFilename) = '.bmp') or
       (ExtractFileExt(ImageFilename) = '.BMP')
    then begin
      InputImage.LoadFromFile(ImageFilename);
      InputImage.PixelFormat := pf24bit;
    end;  // if then
    // Check for and load a Jpeg
    if (ExtractFileExt(ImageFilename) = '.jpg') or
       (ExtractFileExt(ImageFilename) = '.JPG') or
       (ExtractFileExt(ImageFilename) = '.jpeg') or
       (ExtractFileExt(ImageFilename) = '.JPEG') or
       (ExtractFileExt(ImageFilename) = '.jpe') or
       (ExtractFileExt(ImageFilename) = '.JPE')

    then begin
      JPEGLoad := TJPEGImage.Create;
      try   // Protect JPEGImage
        JPEGLoad.LoadFromFile(ImageFilename);
        InputImage.Assign(TBitmap(JPEGLoad));
        InputImage.PixelFormat := pf24bit;
      finally
        JPEGLoad.Free;
      end; // try Protect JPEGImage
    end;  // if then
  except
    on EInvalidGraphic
    do begin
      MessageDlg('The file/s that you have just tried to open may not be valid.'
                + #13
                + #13
                + 'This program can only accept valid bitmap or jpeg files'
                + #13
                + 'with the following file extensions'
                + #13
                + '.bmp  .BMP  .jpg  .JPG  .jpeg  .JPEG  .jpe  .JPE'
                + #13
                + #13
                + 'Please check your files and try again.' ,
                mtError,
                mbOKCancel, 0);
    end;  // do
  end;    // End of EInvalidGraphic Exception handler
end;


procedure DrawControlPoints(Image : TInputImage;
                  var ZoomedImage : TBitmap);
var
DeltaX,
DeltaY  : Integer;
begin
  // Control Point 1
  DeltaX := Round((Image.CP1_X - Image.ViewCentreX) / Image.ScaleFactor) + (ZoomedImage.Width div 2);
  DeltaY := Round((Image.CP1_Y - Image.ViewCentreY) / Image.ScaleFactor) + (ZoomedImage.Height div 2);
  with ZoomedImage.Canvas do
  begin
    Pen.Color := clRed;
    Pen.Width := 1;
    MoveTo(DeltaX - 20, DeltaY);
    LineTo(DeltaX + 20, DeltaY);
    MoveTo(DeltaX, DeltaY - 20);
    LineTo(DeltaX, DeltaY + 20);
    Pen.Color := clRed;
    Pen.Width := 3;
    Arc(DeltaX - 20, DeltaY - 20, DeltaX + 20, DeltaY + 20, DeltaX - 20, DeltaY, DeltaX - 20, DeltaY);
  end;
  // Control Point 2
  DeltaX := Round((Image.CP2_X - Image.ViewCentreX) / Image.ScaleFactor) + (ZoomedImage.Width div 2);
  DeltaY := Round((Image.CP2_Y - Image.ViewCentreY) / Image.ScaleFactor) + (ZoomedImage.Height div 2);
  with ZoomedImage.Canvas do
  begin
    Pen.Color := clBlue;
    Pen.Width := 1;
    MoveTo(DeltaX - 20, DeltaY);
    LineTo(DeltaX + 20, DeltaY);
    MoveTo(DeltaX, DeltaY - 20);
    LineTo(DeltaX, DeltaY + 20);
    Pen.Color := clBlue;
    Pen.Width := 3;
    Arc(DeltaX - 20, DeltaY - 20, DeltaX + 20, DeltaY + 20, DeltaX - 20, DeltaY, DeltaX - 20, DeltaY);
  end;
end;


procedure DrawAlignedControlPoints(var ZoomedImage : TBitmap);
var
DeltaX,
DeltaY  : Integer;
begin
  // Control Point 1
  DeltaX := Round((Form1.MasterCPX - Form1.AlignedCentreX) / Form1.AlignedScale) + (ZoomedImage.Width div 2);
  DeltaY := Round((Form1.MasterCPY - Form1.AlignedCentreY) / Form1.AlignedScale) + (ZoomedImage.Height div 2);
  with ZoomedImage.Canvas do
  begin
    Pen.Color := clRed;
    Pen.Width := 1;
    MoveTo(DeltaX - 20, DeltaY);
    LineTo(DeltaX + 20, DeltaY);
    MoveTo(DeltaX, DeltaY - 20);
    LineTo(DeltaX, DeltaY + 20);
    Pen.Color := clRed;
    Pen.Width := 3;
    Arc(DeltaX - 20, DeltaY - 20, DeltaX + 20, DeltaY + 20, DeltaX - 20, DeltaY, DeltaX - 20, DeltaY);
  end;
  // Control Point 2
  DeltaX := Round((Form1.MasterCP2X - Form1.AlignedCentreX) / Form1.AlignedScale) + (ZoomedImage.Width div 2);
  DeltaY := Round((Form1.MasterCP2Y - Form1.AlignedCentreY) / Form1.AlignedScale) + (ZoomedImage.Height div 2);
  with ZoomedImage.Canvas do
  begin
    Pen.Color := clBlue;
    Pen.Width := 1;
    MoveTo(DeltaX - 20, DeltaY);
    LineTo(DeltaX + 20, DeltaY);
    MoveTo(DeltaX, DeltaY - 20);
    LineTo(DeltaX, DeltaY + 20);
    Pen.Color := clBlue;
    Pen.Width := 3;
    Arc(DeltaX - 20, DeltaY - 20, DeltaX + 20, DeltaY + 20, DeltaX - 20, DeltaY, DeltaX - 20, DeltaY);
  end;
end;


procedure DrawZoomedImage(Image : TInputImage;
                var ZoomedImage : TBitmap);
var
LineIndex       : Integer;
PixelIndex      : Integer;
SourceOriginX   : Integer;
SourceOriginY   : Integer;
SourceX         : Integer;
SourceY         : Integer;
XisInsideImage  : Boolean;
YisInsideImage  : Boolean;
SourceLine      : PByteArray;    // For Scanline
ZoomedLine      : PByteArray;    // For Scanline
BGRed           : Integer;
BGGreen         : Integer;
BGBlue          : Integer;
AdjustedValue   : Integer;
ContrastAdjust  : Extended;
begin
  if Image.Bitmap.Width = 0 then Exit;
  if Image.Bitmap.Height = 0 then Exit;
  XisInsideImage := True;
  YisInsideImage := True;
  BGRed := 0;
  BGGreen := 0;
  BGBlue := 0;
  ContrastAdjust := FindContrastAdjust(Image.GreyAdjust);
  SourceOriginX := Image.ViewCentreX - Round(ZoomedImage.Width * Image.ScaleFactor / 2);
  SourceOriginY := Image.ViewCentreY - Round(ZoomedImage.Height * Image.ScaleFactor / 2);
     SourceLine := Image.Bitmap.ScanLine[0];

    for LineIndex := 0 to ZoomedImage.Height-1 do
    begin
      ZoomedLine := ZoomedImage.ScanLine[LineIndex];
      SourceY := Round(LineIndex*Image.ScaleFactor) + SourceOriginY;
      if (SourceY < 0) or (SourceY > (Image.Bitmap.Height - 1))
        then YisInsideImage := False
        else SourceLine := Image.Bitmap.ScanLine[SourceY];
      for PixelIndex := 0 to ZoomedImage.Width -1 do
      begin
        SourceX := Round(PixelIndex*Image.ScaleFactor) + SourceOriginX;
        if (SourceX < 0) or (SourceX > (Image.Bitmap.Width - 1))
          then XisInsideImage := False;
        if YisInsideImage and XisInsideImage then
        begin
          if Image.GreyAdjust > 0
          then AdjustedValue := Ceil(SourceLine[SourceX * 3] * ContrastAdjust)
          else AdjustedValue := 255 - Ceil((256 - SourceLine[SourceX * 3]) / ContrastAdjust);
          if AdjustedValue > 255 then AdjustedValue := 255;
          if AdjustedValue < 0 then AdjustedValue := 0;
          ZoomedLine[PixelIndex * 3] := AdjustedValue;
          if Image.GreyAdjust > 0
          then AdjustedValue := Ceil(SourceLine[(SourceX * 3) + 1] * ContrastAdjust)
          else AdjustedValue := 255 - Ceil((256 - SourceLine[(SourceX * 3) + 1]) / ContrastAdjust);
          if AdjustedValue > 255 then AdjustedValue := 255;
          if AdjustedValue < 0 then AdjustedValue := 0;
          ZoomedLine[(PixelIndex * 3) + 1] := AdjustedValue;
          if Image.GreyAdjust > 0
          then AdjustedValue := Ceil(SourceLine[(SourceX * 3) + 2] * ContrastAdjust)
          else AdjustedValue := 255 - Ceil((256 - SourceLine[(SourceX * 3) + 2]) / ContrastAdjust);
          if AdjustedValue > 255 then AdjustedValue := 255;
          if AdjustedValue < 0 then AdjustedValue := 0;
          ZoomedLine[(PixelIndex * 3) + 2] := AdjustedValue;
        end
      else
        begin
          ZoomedLine[PixelIndex * 3] := BGRed;
          ZoomedLine[(PixelIndex * 3) + 1] := BGGreen;
          ZoomedLine[(PixelIndex * 3) + 2] := BGBlue;
        end; // if inside image
        XisInsideImage := True;
      end;   // for PixelIndex
      YisInsideImage := True;
    end;     // for LineIndex

  DrawControlPoints(Image, ZoomedImage);

  ZoomedImage.Canvas.Pen.Width := 1;
  ZoomedImage.Canvas.Pen.Color := clBlack;
  ZoomedImage.Canvas.MoveTo(0,0);
  ZoomedImage.Canvas.LineTo(0,ZoomedImage.Height-1);
  ZoomedImage.Canvas.LineTo(ZoomedImage.Width-1,ZoomedImage.Height-1);
  ZoomedImage.Canvas.LineTo(ZoomedImage.Width-1,0);
  ZoomedImage.Canvas.LineTo(0,0);
end;         // DrawZoomedImage


procedure DrawZoomedHistograms(Image : TInputImage;
                    var OutputBitmap : TBitmap);
var
ZoomedHisto : array of Integer;
I : Integer;
PeakValue : Integer;
ZoomedIndex : Integer;
AdjustedIndex : Integer;
DeltaX : Integer;
begin
  try
  // Initialise
  SetLength(ZoomedHisto, OutputBitmap.Width);
  for I := 0 to OutputBitmap.Width - 1
    do ZoomedHisto[I] := 0;
  PeakValue := 1;
  // Convert data to display width
  for I := 0 to (OutputBitmap.Width - 1)
  do begin
    ZoomedIndex := round(I * 254 / (OutputBitmap.Width - 1));
    ZoomedHisto[I] := Image.Histogram[ZoomedIndex];
    if PeakValue < ZoomedHisto[I] then PeakValue := ZoomedHisto[I];
  end;
  // Convert data to display height
  for I := 0 to OutputBitmap.Width - 1
  do begin
    ZoomedHisto[I] := round((ZoomedHisto[I] / PeakValue) * OutputBitmap.Height);
  end;
  // Draw Background
  OutputBitmap.Canvas.Brush.Color := clWhite;
  OutputBitmap.Canvas.FillRect(OutputBitmap.Canvas.ClipRect);
  // Draw Adjusted Histogram
  DeltaX := round(Image.GreyAdjust * OutputBitmap.Width / 254);
  OutputBitmap.Canvas.Pen.Color := clGray;
  OutputBitmap.Canvas.Pen.Width := 1;
  OutputBitmap.Canvas.MoveTo(0, OutputBitmap.Height - 1);
  for I := 0 to OutputBitmap.Width - 1
  do begin
    AdjustedIndex := I - DeltaX;
    if (AdjustedIndex < 0) or (AdjustedIndex > Length(ZoomedHisto) - 1)
    then OutputBitmap.Canvas.LineTo(I, OutputBitmap.Height - 1)
    else begin
      OutputBitmap.Canvas.MoveTo(I, OutputBitmap.Height - 1);
      OutputBitmap.Canvas.LineTo(I, OutputBitmap.Height - 1 - ZoomedHisto[AdjustedIndex]);
    end;
  end;
  // Draw Original Data Line
  OutputBitmap.Canvas.Pen.Color := clBlack;
  OutputBitmap.Canvas.MoveTo(0, OutputBitmap.Height - 1);
  for I := 0 to OutputBitmap.Width - 1
  do begin
      OutputBitmap.Canvas.LineTo(I, OutputBitmap.Height - 1 - ZoomedHisto[I]);
  end;
  // Draw Border
  OutputBitmap.Canvas.Pen.Width := 1;
  OutputBitmap.Canvas.Pen.Color := clBlack;
  OutputBitmap.Canvas.MoveTo(0,0);
  OutputBitmap.Canvas.LineTo(0,OutputBitmap.Height-1);
  OutputBitmap.Canvas.LineTo(OutputBitmap.Width-1,OutputBitmap.Height-1);
  OutputBitmap.Canvas.LineTo(OutputBitmap.Width-1,0);
  OutputBitmap.Canvas.LineTo(0,0);
  // Clean up
  finally
  SetLength(ZoomedHisto, 0);
  end;  // try
end;


procedure DrawAlignedPreview(Image1 : TInputImage;
                             Image2 : TInputImage;
                             Image3 : TInputImage;
                             Master : Integer;
                   var OutputBitmap : TBitmap);
var
ImageCount      : Integer;
Image1Empty,
Image2Empty,
Image3Empty     : Boolean;
OutputX         : Integer;
OutputY         : Integer;
MasterOriginX   : Integer;
MasterOriginY   : Integer;
MasterX         : Integer;
MasterY         : Integer;
WorldMasterX    : Double;
WorldMasterY    : Double;
MasterAzimuth   : Extended;
MasterMeridian  : Extended;
SourceAzimuth   : Extended;
SourceMeridian  : Extended;
TempLength      : Extended;
TempAngle       : Extended;
SourceX         : Integer;
SourceY         : Integer;
WorldSourceX    : Double;
WorldSourceY    : Double;
SourceLine1     : PByteArray;    // For Scanline
SourceLine2     : PByteArray;    // For Scanline
SourceLine3     : PByteArray;    // For Scanline
AlignedLine     : PByteArray;    // For Scanline
SourceLines1    : array of PByteArray;
SourceLines2    : array of PByteArray;
SourceLines3    : array of PByteArray;
A               : Extended;
Sin_1           : Extended;
Cos_1           : Extended;
Sin_2           : Extended;
Cos_2           : Extended;
Sin_3           : Extended;
Cos_3           : Extended;
AdjustedValue   : Integer;
R     : Integer;
G     : Integer;
B     : Integer;
UpdateInterval     : Integer;
SaveCursor    :TCursor;
OutputBorderTop    : Integer;
OutputBorderBottom : Integer;
OutputBorderLeft   : Integer;
OutputBorderRight  : Integer;
Border             : Integer;
Image1ContrastAdjust : Extended;
Image2ContrastAdjust : Extended;
Image3ContrastAdjust : Extended;
begin
  ImageCount := 0;
  TempLength := 0;
  TempAngle  := 0;
  Image1Empty := False;
  Image2Empty := False;
  Image3Empty := False;
  if Image1.FullFilename = '' then Image1Empty := True else Inc(ImageCount);
  if Image2.FullFilename = '' then Image2Empty := True else Inc(ImageCount);
  if Image3.FullFilename = '' then Image3Empty := True else Inc(ImageCount);
  if ImageCount = 0 then Exit;

  Image1ContrastAdjust := FindContrastAdjust(Image1.GreyAdjust);
  Image2ContrastAdjust := FindContrastAdjust(Image2.GreyAdjust);
  Image3ContrastAdjust := FindContrastAdjust(Image3.GreyAdjust);

  SaveCursor := Screen.Cursor;
  if Form1.LivePreviewCheckBox.Checked = False
    then Screen.Cursor := crHourGlass;    { Show hourglass cursor }
  try          // Protect Screen Cursor
    // Build Source Scanlines Arrays
    if not(Image1Empty)
    then begin
      SetLength(SourceLines1, Image1.Bitmap.Height);
      for OutputY := 0 to Image1.Bitmap.Height - 1 do
      begin
        SourceLines1[OutputY] := Image1.Bitmap.ScanLine[OutputY];
      end;
    end;
    if not(Image2Empty)
    then begin
      SetLength(SourceLines2, Image2.Bitmap.Height);
      for OutputY := 0 to Image2.Bitmap.Height - 1 do
      begin
        SourceLines2[OutputY] := Image2.Bitmap.ScanLine[OutputY];
      end;
    end;
    if not(Image3Empty)
    then begin
      SetLength(SourceLines3, Image3.Bitmap.Height);
      for OutputY := 0 to Image3.Bitmap.Height - 1 do
      begin
        SourceLines3[OutputY] := Image3.Bitmap.ScanLine[OutputY];
      end;
    end;

    SinCos(Image1.Rotation, Sin_1, Cos_1);
    SinCos(Image2.Rotation, Sin_2, Cos_2);
    SinCos(Image3.Rotation, Sin_3, Cos_3);
    MasterOriginX := Form1.AlignedCentreX - Round(OutputBitmap.Width * Form1.AlignedScale / 2);
    MasterOriginY := Form1.AlignedCentreY - Round(OutputBitmap.Height * Form1.AlignedScale / 2);

    UpdateInterval := 1 + (OutputBitmap.Height div 20);

    for OutputY := 0 to OutputBitmap.Height - 1 do
    begin
      AlignedLine := OutputBitmap.ScanLine[OutputY];
      if Form1.FOV > 0
      then MasterY := MasterOriginY + Round(OutputY * Form1.AlignedScale) - Form1.MasterCentreY
      else MasterY := MasterOriginY + Round(OutputY * Form1.AlignedScale) - Form1.MasterCPY;

      for OutputX := 0 to OutputBitmap.Width - 1 do
      begin
        R := 0;
        G := 0;
        B := 0;
        if Form1.FOV > 0
        then begin  //  FOV > 0
          MasterX := MasterOriginX + Round(OutputX * Form1.AlignedScale) - Form1.MasterCentreX;
          ImageToSpherical(MasterX, MasterY, MasterAzimuth, MasterMeridian);
          // Find spherical distance from master point to control point 1      (1)
          TempLength := SphericalLength(Form1.MasterCP_Azimuth, MasterAzimuth, (MasterMeridian - Form1.MasterCP_Meridian));
          // Find spherical direction of master point from control point 1     (2)
          TempAngle := SphericalAngleImageStage_1(MasterAzimuth, TempLength, Form1.MasterCP_Azimuth, (MasterMeridian - Form1.MasterCP_Meridian));
        end         //  FOV > 0
        else begin  //  FOV = 0
          MasterX := MasterOriginX + Round(OutputX * Form1.AlignedScale) - Form1.MasterCPX;
        end;        //  FOV = 0

        if not(Image1Empty) then
        begin
          if Master = 1
          then begin
            if Form1.FOV > 0
            then begin  //  FOV > 0
              SourceX := MasterX + Image1.CentreX;
              SourceY := MasterY + Image1.CentreY;
            end         //  FOV > 0
            else begin  //  FOV = 0
              SourceX := MasterX + Image1.CP1_X;
              SourceY := MasterY + Image1.CP1_Y;
            end         //  FOV = 0
          end           //  if Master
          else begin
            if Form1.FOV > 0
            then begin  //  FOV > 0
              TempLength := TempLength * Image1.Scale;
              //  (3)
              SourceAzimuth := SphericalLength(TempLength, Image1.CP1_Azimuth, (TempAngle - Image1.Rotation));
              //  (4)
              SourceMeridian := SphericalAngleImageStage_2(TempLength, SourceAzimuth, Image1.CP1_Azimuth, (TempAngle - Image1.Rotation));

              SourceMeridian := SourceMeridian + Image1.CP1_Meridian;

              SphericalToImage(SourceAzimuth, SourceMeridian, SourceX, SourceY);

              SourceX := SourceX + Image1.CentreX;
              SourceY := SourceY + Image1.CentreY;

              TempLength := TempLength / Image1.Scale;

              if Abs(SourceMeridian) > pi / 2
              then begin
                SourceX := -1;
                SourceY := -1;
              end;

            end         //  FOV > 0
            else begin  //  FOV = 0
              SourceY := Round(Image1.CP1_Y
                            + (Image1.Scale
                            * ((MasterX * Sin_1) + (MasterY * Cos_1))));
              SourceX := Round(Image1.CP1_X
                            + (Image1.Scale
                            * ((MasterX * Cos_1) - (MasterY * Sin_1))));
            end;        //  FOV = 0
          end;          //  if Master = 1

          if (SourceY >= 0) and (SourceY <= (Image1.Bitmap.Height - 1)) and
             (SourceX >= 0) and (SourceX <= (Image1.Bitmap.Width - 1)) then
          begin
            SourceLine1 := SourceLines1[SourceY];
            AdjustedValue := Ceil(sourceline1[(SourceX * 3)] * Image1ContrastAdjust);
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
            R := R + AdjustedValue;
            AdjustedValue := Ceil(sourceline1[(SourceX * 3) + 1] * Image1ContrastAdjust);
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
            G := G + AdjustedValue;
            AdjustedValue := Ceil(sourceline1[(SourceX * 3) + 2] * Image1ContrastAdjust);
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
            B := B + AdjustedValue;
          end
        end;  // if inside image

        if not(Image2Empty) then
        begin

          if Master = 2
          then begin
            if Form1.FOV > 0
            then begin  //  FOV > 0
              SourceX := MasterX + Image2.CentreX;
              SourceY := MasterY + Image2.CentreY;
            end         //  FOV > 0
            else begin  //  FOV = 0
              SourceX := MasterX + Image2.CP1_X;
              SourceY := MasterY + Image2.CP1_Y;
            end         //  FOV = 0
          end
          else begin
            if Form1.FOV > 0
            then begin  //  FOV > 0
              TempLength := TempLength * Image2.Scale;
              SourceAzimuth := SphericalLength(TempLength, Image2.CP1_Azimuth, (TempAngle - Image2.Rotation));
              SourceMeridian := SphericalAngleImageStage_2(TempLength, SourceAzimuth, Image2.CP1_Azimuth, (TempAngle - Image2.Rotation));

              SourceMeridian := SourceMeridian + Image2.CP1_Meridian;

              SphericalToImage(SourceAzimuth, SourceMeridian, SourceX, SourceY);

              SourceX := SourceX + Image2.CentreX;
              SourceY := SourceY + Image2.CentreY;

              TempLength := TempLength / Image2.Scale;

              if Abs(SourceMeridian) > pi / 2
              then begin
                SourceX := -1;
                SourceY := -1;
              end;

            end         //  FOV > 0
            else begin  //  FOV = 0
              SourceY := Round(Image2.CP1_Y
                            + (Image2.Scale
                            * ((MasterX * Sin_2) + (MasterY * Cos_2))));
              SourceX := Round(Image2.CP1_X
                            + (Image2.Scale
                            * ((MasterX * Cos_2) - (MasterY * Sin_2))));
            end;        //  FOV = 0
          end;          //  if Master = 2

          if (SourceY >= 0) and (SourceY <= (Image2.Bitmap.Height - 1)) and
             (SourceX >= 0) and (SourceX <= (Image2.Bitmap.Width - 1)) then
          begin
            SourceLine2 := SourceLines2[SourceY];
            AdjustedValue := Ceil(sourceline2[(SourceX * 3)] * Image2ContrastAdjust);
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
            R := R + AdjustedValue;
            AdjustedValue := Ceil(sourceline2[(SourceX * 3) + 1] * Image2ContrastAdjust);
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
            G := G + AdjustedValue;
            AdjustedValue := Ceil(sourceline2[(SourceX * 3) + 2] * Image2ContrastAdjust);
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
            B := B + AdjustedValue;
          end
        end;  // if inside image

        if not(Image3Empty) then
        begin

          if Master = 3
          then begin
            if Form1.FOV > 0
            then begin  //  FOV > 0
              SourceX := MasterX + Image3.CentreX;
              SourceY := MasterY + Image3.CentreY;
            end         //  FOV > 0
            else begin  //  FOV = 0
              SourceX := MasterX + Image3.CP1_X;
              SourceY := MasterY + Image3.CP1_Y;
            end         //  FOV = 0
          end
          else begin
            if Form1.FOV > 0
            then begin  //  FOV > 0
              TempLength := TempLength * Image3.Scale;
              SourceAzimuth := SphericalLength(TempLength, Image3.CP1_Azimuth, (TempAngle - Image3.Rotation));
              SourceMeridian := SphericalAngleImageStage_2(TempLength, SourceAzimuth, Image3.CP1_Azimuth, (TempAngle - Image3.Rotation));

              SourceMeridian := SourceMeridian + Image3.CP1_Meridian;

              SphericalToImage(SourceAzimuth, SourceMeridian, SourceX, SourceY);

              SourceX := SourceX + Image3.CentreX;
              SourceY := SourceY + Image3.CentreY;

              TempLength := TempLength / Image3.Scale;
              if Abs(SourceMeridian) > pi / 2
              then begin
                SourceX := -1;
                SourceY := -1;
              end;

            end         //  FOV > 0
            else begin  //  FOV = 0
              SourceY := Round(Image3.CP1_Y
                            + (Image3.Scale
                            * ((MasterX * Sin_3) + (MasterY * Cos_3))));
              SourceX := Round(Image3.CP1_X
                            + (Image3.Scale
                            * ((MasterX * Cos_3) - (MasterY * Sin_3))));
            end;        //  FOV = 0
          end;          //  if Master = 3

          if (SourceY >= 0) and (SourceY <= (Image3.Bitmap.Height - 1)) and
             (SourceX >= 0) and (SourceX <= (Image3.Bitmap.Width - 1)) then
          begin
            SourceLine3 := SourceLines3[SourceY];
            AdjustedValue := Ceil(sourceline3[(SourceX * 3)] * Image3ContrastAdjust);
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
            R := R + AdjustedValue;
            AdjustedValue := Ceil(sourceline3[(SourceX * 3) + 1] * Image3ContrastAdjust);
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
            G := G + AdjustedValue;
            AdjustedValue := Ceil(sourceline3[(SourceX * 3) + 2] * Image3ContrastAdjust);
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
            B := B + AdjustedValue;
          end
        end;  // if inside image

        AdjustedValue := R div ImageCount;
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
        AlignedLine[OutputX * 3] := AdjustedValue;
        AdjustedValue := G div ImageCount;
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
        AlignedLine[(OutputX * 3) + 1] := AdjustedValue;
        AdjustedValue := B div ImageCount;
            if AdjustedValue > 255 then AdjustedValue := 255;
            if AdjustedValue < 0 then AdjustedValue := 0;
        AlignedLine[(OutputX * 3) + 2] := AdjustedValue;
      end;  // for PixelIndex

      if OutputY mod UpdateInterval = 0
      then begin
        Form1.MainProgressBar.Position := round(Form1.MainProgressBar.Max * OutputY / OutputBitmap.Height);
        Form1.StatusBar1.Repaint;
        Form1.StatusBar1.Panels[1].Text := 'Processing';
      end;

    end;  // for LineIndex

    // Put the Control Points on the image
    // Work out which image is the master
    if Form1.Master = 1 then DrawAlignedControlPoints(OutputBitmap);
    if Form1.Master = 2 then DrawAlignedControlPoints(OutputBitmap);
    if Form1.Master = 3 then DrawAlignedControlPoints(OutputBitmap);

    // Draw Output Image Border
    OutputBorderTop := (OutputBitmap.Height div 2) - Round(Form1.AlignedCentreY / Form1.AlignedScale);
    OutputBorderBottom := (OutputBitmap.Height div 2) + Round((Form1.MasterHeight - Form1.AlignedCentreY) / Form1.AlignedScale);
    OutputBorderLeft := (OutputBitmap.Width div 2) - Round(Form1.AlignedCentreX / Form1.AlignedScale);
    OutputBorderRight := (OutputBitmap.Width div 2) + Round((Form1.MasterWidth - Form1.AlignedCentreX) / Form1.AlignedScale);
    if SaveForm.BordersCheckBox.Checked
    then begin
      Border := Round(StrToIntDef(SaveForm.TopBdrEdit.Text, 0) / Form1.AlignedScale);
      OutputBorderTop := OutputBorderTop - Border;
      Border := Round(StrToIntDef(SaveForm.BottomBdrEdit.Text, 0) / Form1.AlignedScale);
      OutputBorderBottom := OutputBorderBottom + Border;
      Border := Round(StrToIntDef(SaveForm.LeftBdrEdit.Text, 0) / Form1.AlignedScale);
      OutputBorderLeft := OutputBorderLeft - Border;
      Border := Round(StrToIntDef(SaveForm.RightBdrEdit.Text, 0) / Form1.AlignedScale);
      OutputBorderRight := OutputBorderRight + Border;
    end;
    OutputBitmap.Canvas.Pen.Width := 1;
    OutputBitmap.Canvas.Pen.Color := clGray;
    OutputBitmap.Canvas.MoveTo(OutputBorderLeft,OutputBorderTop);
    OutputBitmap.Canvas.LineTo(OutputBorderLeft,OutputBorderBottom);
    OutputBitmap.Canvas.LineTo(OutputBorderRight,OutputBorderBottom);
    OutputBitmap.Canvas.LineTo(OutputBorderRight,OutputBorderTop);
    OutputBitmap.Canvas.LineTo(OutputBorderLeft,OutputBorderTop);

    // Draw Border
    OutputBitmap.Canvas.Pen.Width := 1;
    OutputBitmap.Canvas.Pen.Color := clBlack;
    OutputBitmap.Canvas.MoveTo(0,0);
    OutputBitmap.Canvas.LineTo(0,OutputBitmap.Height-1);
    OutputBitmap.Canvas.LineTo(OutputBitmap.Width-1,OutputBitmap.Height-1);
    OutputBitmap.Canvas.LineTo(OutputBitmap.Width-1,0);
    OutputBitmap.Canvas.LineTo(0,0);

    Form1.MainProgressBar.Position := 0;
    Form1.StatusBar1.Repaint;
    Form1.StatusBar1.Panels[1].Text := 'Ready';

  finally
    Screen.Cursor := SaveCursor;  { Always restore to normal }
    SetLength(SourceLines1, 0);
    SetLength(SourceLines2, 0);
    SetLength(SourceLines3, 0);
  end;      // try protect screen cursor

end;


procedure BuildAlignedBitmap(Source   : TInputImage;
                             IsMaster : Boolean;
                         var Result   : TBitmap);
var
LineIndex       : Integer;
PixelIndex      : Integer;
MasterAzimuth   : Extended;
MasterMeridian  : Extended;
SourceAzimuth   : Extended;
SourceMeridian  : Extended;
TempLength      : Extended;
TempAngle       : Extended;
SourceX         : Extended;
SourceY         : Extended;
SourceX_Int     : Integer;
SourceY_Int     : Integer;
SourceLine1     : PByteArray;    // For Scanline
SourceLine2     : PByteArray;    // For Scanline
ResultLine      : PByteArray;    // For Scanline
SourceLines     : array of PByteArray;
A               : Extended;
sinA            : Extended;
cosA            : Extended;
X_cp            : Extended;
Y_cp            : Extended;
DeltaX          : Extended;
DeltaY          : Extended;
NewValue        : Extended;
NewValue_Int    : Integer;
UpdateInterval  : Integer;
TopBdr          : Integer;
BottomBdr       : Integer;
LeftBdr         : Integer;
RightBdr        : Integer;
begin
  // Build Source Scanline Array
  try
    SetLength(SourceLines, Source.Bitmap.Height);
    for LineIndex := 0 to Source.Bitmap.Height - 1 do
    begin
      SourceLines[LineIndex] := Source.Bitmap.ScanLine[LineIndex];
    end;

    A := Source.Rotation;
    SinCos(A, sinA, cosA);

    UpdateInterval := Result.Height div 50;
    if UpdateInterval = 0 then UpdateInterval := 1;

    if SaveForm.BordersCheckBox.Checked
    then begin
      TopBdr    := StrToIntDef(SaveForm.TopBdrEdit.Text,0);
      BottomBdr := StrToIntDef(SaveForm.BottomBdrEdit.Text,0);
      LeftBdr   := StrToIntDef(SaveForm.LeftBdrEdit.Text,0);
      RightBdr  := StrToIntDef(SaveForm.RightBdrEdit.Text,0);
    end
    else begin
      TopBdr    := 0;
      BottomBdr := 0;
      LeftBdr   := 0;
      RightBdr  := 0;
    end;

    for LineIndex := 0 to Result.Height - 1 do
    begin
      ResultLine := Result.ScanLine[LineIndex];

      if Form1.FOV > 0
        then Y_cp := LineIndex - (Form1.MasterHeight / 2) - TopBdr + 0.5
        else Y_cp := LineIndex - Form1.MasterCPY {- TopBdr};

      for PixelIndex := 0 to Result.Width - 1 do
      begin

        if IsMaster
        then begin
          if Form1.FOV > 0
          then begin  //  FOV > 0
            SourceX := PixelIndex - LeftBdr + 0.5;
            SourceY := LineIndex - TopBdr + 0.5;
          end         //  FOV > 0
          else begin  //  FOV = 0
            SourceX := PixelIndex - LeftBdr + 0.5;
            SourceY := LineIndex - TopBdr + 0.5;
          end         //  FOV = 0
        end           //  if IsMaster
        else begin    //  if not IsMaster

          if Form1.FOV > 0
          then begin  //  FOV > 0
            X_cp := PixelIndex - (Form1.MasterWidth / 2) - LeftBdr + 0.5;
            RealImageToSpherical(X_cp, Y_cp, MasterAzimuth, MasterMeridian);
            // Find spherical distance from master point to control point 1      (1)
            TempLength := SphericalLength(Form1.MasterCP_Azimuth, MasterAzimuth, (MasterMeridian - Form1.MasterCP_Meridian));
            // Find spherical direction of master point from control point 1     (2)
            TempAngle := SphericalAngleImageStage_1(MasterAzimuth, TempLength, Form1.MasterCP_Azimuth, (MasterMeridian - Form1.MasterCP_Meridian));
            TempLength := TempLength * Source.Scale;
            // Find the azimuth of the point in the source image                 (3)
            SourceAzimuth := SphericalLength(TempLength, Source.CP1_Azimuth, (TempAngle - Source.Rotation));
            // Find the relative meridian of the point in the source image       (4)
            SourceMeridian := SphericalAngleImageStage_2(TempLength, SourceAzimuth, Source.CP1_Azimuth, (TempAngle - Source.Rotation));
            // Find the absolute meridian of the point in the source image
            SourceMeridian := SourceMeridian + Source.CP1_Meridian;
            // Translate these coords back into relative source image coords
            SphericalToRealImage(SourceAzimuth, SourceMeridian, SourceX, SourceY);
            // Convert these into absolute image coords
            SourceX := SourceX + (Source.Bitmap.Width  / 2);
            SourceY := SourceY + (Source.Bitmap.Height / 2);
            // Check meridian has not wraped around to the back of the sphere
            if Abs(SourceMeridian) > pi / 2
            then begin
              // And set the pixel to be off the image if it is around the back
              SourceX := -1;
              SourceY := -1;
            end;
          end         //  FOV > 0
          else begin  //  FOV = 0
            X_cp    := PixelIndex - Form1.MasterCPX {- LeftBdr};
            SourceX := Source.CP1_X + 0.5 + (Source.Scale * ((X_cp * CosA) - (Y_cp * sinA)));
            SourceY := Source.CP1_Y + 0.5 + (Source.Scale * ((X_cp * sinA) + (Y_cp * CosA)));
          end;        //  FOV = 0

        end;          //  if not Master

        // Range Check
        if SourceX >  MaxInt then SourceX :=  MaxInt;
        if SourceX < -MaxInt then SourceX := -MaxInt;
        if SourceY >  MaxInt then SourceY :=  MaxInt;
        if SourceY < -MaxInt then SourceY := -MaxInt;
        SourceX := RoundTo(SourceX, -5);
        SourceY := RoundTo(SourceY, -5);

        if (SourceX >= 0.5) and (SourceX <= Source.Bitmap.Width  - 0.5) and
           (SourceY >= 0.5) and (SourceY <= Source.Bitmap.Height - 0.5)
        then begin    // if pixel IS inside original image
          // Assign pixel from original image to the output bitmap
          if SourceX > 1 then SourceX := SourceX - 0.00001;   // to prevent running off
          if SourceY > 1 then SourceY := SourceY - 0.00001;   // the end of the image
          SourceX_Int  := Floor(SourceX - 0.5);
          SourceY_Int  := Floor(SourceY - 0.5);
          SourceLine1  := SourceLines[SourceY_Int];
          SourceLine2  := SourceLines[SourceY_Int + 1];
          DeltaX       := SourceX - SourceX_Int;
          DeltaY       := SourceY - SourceY_Int;
          NewValue     := Interpolate(SourceLine1[SourceX_Int * 3],
                                      SourceLine1[(SourceX_Int + 1) * 3],
                                      SourceLine2[SourceX_Int * 3],
                                      SourceLine2[(SourceX_Int + 1) * 3],
                                      DeltaX, DeltaY);
          NewValue_Int := CapPixelValue(NewValue);
          ResultLine[PixelIndex * 3]       := NewValue_Int;
          NewValue     := Interpolate(SourceLine1[(SourceX_Int * 3) + 1],
                                      SourceLine1[((SourceX_Int + 1) * 3) + 1],
                                      SourceLine2[(SourceX_Int * 3) + 1],
                                      SourceLine2[((SourceX_Int + 1) * 3) + 1],
                                      DeltaX, DeltaY);
          NewValue_Int := CapPixelValue(NewValue);
          ResultLine[(PixelIndex * 3) + 1] := NewValue_Int;
          NewValue     := Interpolate(SourceLine1[(SourceX_Int * 3) + 2],
                                      SourceLine1[((SourceX_Int + 1) * 3) + 2],
                                      SourceLine2[(SourceX_Int * 3) + 2],
                                      SourceLine2[((SourceX_Int + 1) * 3) + 2],
                                      DeltaX, DeltaY);
          NewValue_Int := CapPixelValue(NewValue);
          ResultLine[(PixelIndex * 3) + 2] := NewValue_Int;
        end
        else begin    // if pixel IS NOT inside original image
          ResultLine[PixelIndex * 3]       := Source.GreyPoint;
          ResultLine[(PixelIndex * 3) + 1] := Source.GreyPoint;
          ResultLine[(PixelIndex * 3) + 2] := Source.GreyPoint;
        end           //  if inside original image
      end;            //  for PixelIndex

      if LineIndex mod UpdateInterval = 0
      then begin
        Form1.MainProgressBar.Position := round(Form1.MainProgressBar.Max * LineIndex / Result.Height);
        Form1.StatusBar1.Invalidate;
        Application.ProcessMessages;
      end;

    end;              //  for LineIndex
  finally
    SetLength(SourceLines, 0);
  end;                //  try
end;


procedure DrawQuickPreview(DeltaX        : Integer;
                           DeltaY       : Integer;
                           InputPreview : TBitmap;
                           var Result   : TBitmap);
begin
  Result.Width := InputPreview.Width;
  Result.Height := InputPreview.Height;
  Result.PixelFormat := pf24bit;
  Result.Canvas.Brush.Color := clBlack;
  Result.Canvas.Rectangle(0, 0, Result.Width - 1, Result.Height - 1);
  Result.Canvas.Draw(DeltaX, DeltaY, InputPreview);
  Result.Canvas.Pen.Width := 1;
  Result.Canvas.Pen.Color := clBlack;
  Result.Canvas.MoveTo(0,0);
  Result.Canvas.LineTo(0,Result.Height-1);
  Result.Canvas.LineTo(Result.Width-1,Result.Height-1);
  Result.Canvas.LineTo(Result.Width-1,0);
  Result.Canvas.LineTo(0,0);
end;


function Interpolate(A_PixelValue  : Byte;
                     B_PixelValue  : Byte;
                     C_PixelValue  : Byte;
                     D_PixelValue  : Byte;
                     X             : Extended;
                     Y             : Extended): Extended;
begin
  // Nearest Neighbor

  // Bi-Linear interpolation based on 'area'
  Result := ((1.5 - X) * (1.5 - Y) * A_PixelValue) +
            ((X - 0.5) * (1.5 - Y) * B_PixelValue) +
            ((1.5 - X) * (Y - 0.5) * C_PixelValue) +
            ((X - 0.5) * (Y - 0.5) * D_PixelValue);
end;



function CapPixelValue(Value : Extended): Word;
var
Value_Int : Integer;
begin
  Value_Int := Round(Value);
  If Value_Int < 0   then Value_Int := 0;
  If Value_Int > 255 then Value_Int := 255;
  Result := Value_Int;
end;



end.
