
  /////////////////////////////////////////////////////
  /////////////////////////////////////////////////////
  ////                                             ////
  ////        HDR Alignment Tool Sourcefile        ////
  ////         Free for non-commercial use         ////
  ////                                             ////
  ////        Copyright  2007 Boyd Eagleson       ////
  ////    (Contact... 'Eggy Boil' on Flickr.com)   ////
  ////                                             ////
  /////////////////////////////////////////////////////
  /////////////////////////////////////////////////////

unit SaveUnit;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, Buttons, ComCtrls, ExtCtrls, Math;

type
  TSaveForm = class(TForm)
    OKBtn: TBitBtn;
    JPGRadio: TRadioButton;
    BMPRadio: TRadioButton;
    Label2: TLabel;
    JPGQualitySlider: TTrackBar;
    JPGQualityLabel: TLabel;
    Image1SourceNameEdit: TEdit;
    Image2SourceNameEdit: TEdit;
    Image3SourceNameEdit: TEdit;
    AlignedTextEdit: TEdit;
    CancelBtn: TBitBtn;
    ImagePreviewPaintBox: TPaintBox;
    OutputFileList: TListBox;
    SelectDirBtn: TBitBtn;
    RecalculateBtn: TButton;
    Label1: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    OutputDirLabel: TLabel;
    Label6: TLabel;
    AppendCheckBox: TCheckBox;
    ChooseDirectoryDialog: TOpenDialog;
    FileOverwriteWarningStaticText: TStaticText;
    OutputDirEdit: TEdit;
    Image1FileOverwriteMarker: TShape;
    Image2FileOverwriteMarker: TShape;
    Image3FileOverwriteMarker: TShape;
    BordersCheckBox: TCheckBox;
    TopBdrEdit: TEdit;
    LeftBdrEdit: TEdit;
    BottomBdrEdit: TEdit;
    RightBdrEdit: TEdit;
    OutputFullDirLabel: TLabel;
    ProgressBar: TProgressBar;
    Shape1: TShape;
    Image1MasterMarker: TShape;
    Image2MasterMarker: TShape;
    Image3MasterMarker: TShape;
    procedure OKBtnClick(Sender: TObject);
    procedure JPGQualitySliderChange(Sender: TObject);
    procedure CancelBtnClick(Sender: TObject);
    procedure SelectDirBtnClick(Sender: TObject);
    procedure RecalculateBtnClick(Sender: TObject);
    procedure OutputFileListClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure OutputDirEditChange(Sender: TObject);
    procedure ImagePreviewPaintBoxPaint(Sender: TObject);
    procedure TopBdrEditChange(Sender: TObject);
    procedure BordersCheckBoxClick(Sender: TObject);
    procedure Image1SourceNameEditClick(Sender: TObject);
    procedure Image2SourceNameEditClick(Sender: TObject);
    procedure Image3SourceNameEditClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    Image1FileName       : String;
    Image2FileName       : String;
    Image3FileName       : String;
    Image1OutputFileName : String;
    Image2OutputFileName : String;
    Image3OutputFileName : String;
    Image1FullOutputPath : String;
    Image2FullOutputPath : String;
    Image3FullOutputPath : String;
    OutputDirectory      : String;
    Result               : TBitmap;
    SelectedImage        : Integer;
  end;

var
  SaveForm: TSaveForm;

implementation

uses InterfaceUnit, ImageRoutines;

{$R *.dfm}



procedure BuildSavePreviewBitmap(Source   : TInputImage;
                                 IsMaster : Boolean;
                             var Result   : TBitmap);
var
LineIndex       : Integer;
PixelIndex      : Integer;
UpdateInterval  : Integer;
SaveCursor      : TCursor;
PreviewZoom     : Extended;
TempZoom        : Extended;
MasterAzimuth   : Extended;
MasterMeridian  : Extended;
SourceAzimuth   : Extended;
SourceMeridian  : Extended;
TempLength      : Extended;
TempAngle       : Extended;
SourceX         : Extended;
SourceY         : Extended;
SourceX_Int     : Integer;
SourceY_Int     : Integer;
SourceLine1     : PByteArray;    // For Scanline
SourceLine2     : PByteArray;    // For Scanline
ResultLine      : PByteArray;    // For Scanline
SourceLines     : array of PByteArray;
A               : Extended;
sinA            : Extended;
cosA            : Extended;
X_cp            : Extended;
Y_cp            : Extended;
DeltaX          : Extended;
DeltaY          : Extended;
NewValue        : Extended;
NewValue_Int    : Integer;
TopBdr          : Integer;
BottomBdr       : Integer;
LeftBdr         : Integer;
RightBdr        : Integer;
TempResult      : TBitmap;
X_Offset        : Integer;
Y_Offset        : Integer;
begin

  TempResult := TBitmap.Create;

  SaveCursor := Screen.Cursor;
  Screen.Cursor := crHourGlass;    { Show hourglass cursor }

  try

    // Build Source Scanline Array
    SetLength(SourceLines, Source.Bitmap.Height);
    for LineIndex := 0 to Source.Bitmap.Height - 1 do
    begin
      SourceLines[LineIndex] := Source.Bitmap.ScanLine[LineIndex];
    end;

    A := Source.Rotation;
    SinCos(A, sinA, cosA);

    if SaveForm.BordersCheckBox.Checked
    then begin
      TopBdr    := StrToIntDef(SaveForm.TopBdrEdit.Text,0);
      BottomBdr := StrToIntDef(SaveForm.BottomBdrEdit.Text,0);
      LeftBdr   := StrToIntDef(SaveForm.LeftBdrEdit.Text,0);
      RightBdr  := StrToIntDef(SaveForm.RightBdrEdit.Text,0);
    end
    else begin
      TopBdr    := 0;
      BottomBdr := 0;
      LeftBdr   := 0;
      RightBdr  := 0;
    end;

    if ((Form1.MasterHeight + TopBdr + BottomBdr) / Result.Height) > ((Form1.MasterWidth + LeftBdr + RightBdr) / Result.Width)
    then PreviewZoom := 1.1 * (Form1.MasterHeight + TopBdr + BottomBdr) / Result.Height
    else PreviewZoom := 1.1 * (Form1.MasterWidth + LeftBdr + RightBdr) / Result.Width;

    TempZoom := 1;
    if PreviewZoom < 1
    then begin
      TempZoom := PreviewZoom;
      PreviewZoom := 1;
    end;

    TempResult.PixelFormat := pf24bit;
    TempResult.Width  := Ceil((Form1.MasterWidth + LeftBdr + RightBdr)  / PreviewZoom);
    TempResult.Height := Ceil((Form1.MasterHeight + TopBdr + BottomBdr) / PreviewZoom);

    UpdateInterval := TempResult.Height div 50;
    if UpdateInterval = 0 then UpdateInterval := 1;

    for LineIndex := 0 to TempResult.Height - 1 do
    begin
      ResultLine := TempResult.ScanLine[LineIndex];

      if Form1.FOV > 0
        then Y_cp := (LineIndex * PreviewZoom) - (Form1.MasterHeight / 2) - TopBdr + 0.5
        else Y_cp := (LineIndex * PreviewZoom) - Form1.MasterCPY - TopBdr;

      for PixelIndex := 0 to TempResult.Width - 1 do
      begin

        if IsMaster
        then begin
          if Form1.FOV > 0
          then begin  //  FOV > 0
            SourceX := (PixelIndex * PreviewZoom) - LeftBdr + 0.5;
            SourceY := (LineIndex * PreviewZoom) - TopBdr + 0.5;
          end         //  FOV > 0
          else begin  //  FOV = 0
            SourceX := (PixelIndex * PreviewZoom) - LeftBdr + 0.5;
            SourceY := (LineIndex * PreviewZoom) - TopBdr + 0.5;
          end         //  FOV = 0
        end           //  if IsMaster
        else begin    //  if not IsMaster

          if Form1.FOV > 0
          then begin  //  FOV > 0
            X_cp := (PixelIndex * PreviewZoom) - (Form1.MasterWidth / 2) - LeftBdr + 0.5;
            RealImageToSpherical(X_cp, Y_cp, MasterAzimuth, MasterMeridian);
            // Find spherical distance from master point to control point 1      (1)
            TempLength := SphericalLength(Form1.MasterCP_Azimuth, MasterAzimuth, (MasterMeridian - Form1.MasterCP_Meridian));
            // Find spherical direction of master point from control point 1     (2)
            TempAngle := SphericalAngleImageStage_1(MasterAzimuth, TempLength, Form1.MasterCP_Azimuth, (MasterMeridian - Form1.MasterCP_Meridian));
            TempLength := TempLength * Source.Scale;
            // Find the azimuth of the point in the source image                 (3)
            SourceAzimuth := SphericalLength(TempLength, Source.CP1_Azimuth, (TempAngle - Source.Rotation));
            // Find the relative meridian of the point in the source image       (4)
            SourceMeridian := SphericalAngleImageStage_2(TempLength, SourceAzimuth, Source.CP1_Azimuth, (TempAngle - Source.Rotation));
            // Find the absolute meridian of the point in the source image
            SourceMeridian := SourceMeridian + Source.CP1_Meridian;
            // Translate these coords back into relative source image coords
            SphericalToRealImage(SourceAzimuth, SourceMeridian, SourceX, SourceY);
            // Convert these into absolute image coords
            SourceX := SourceX + (Source.Bitmap.Width  / 2);
            SourceY := SourceY + (Source.Bitmap.Height / 2);
            // Check meridian has not wraped around to the back of the sphere
            if Abs(SourceMeridian) > pi / 2
            then begin
              // And set the pixel to be off the image if it is around the back
              SourceX := -1;
              SourceY := -1;
            end;
          end         //  FOV > 0
          else begin  //  FOV = 0
            X_cp    := (PixelIndex * PreviewZoom) - Form1.MasterCPX - LeftBdr;
            SourceX := Source.CP1_X + 0.5 + (Source.Scale * ((X_cp * CosA) - (Y_cp * sinA)));
            SourceY := Source.CP1_Y + 0.5 + (Source.Scale * ((X_cp * sinA) + (Y_cp * CosA)));
          end;        //  FOV = 0

        end;          //  if not Master

        // Range Check
        if SourceX >  MaxInt then SourceX :=  MaxInt;
        if SourceX < -MaxInt then SourceX := -MaxInt;
        if SourceY >  MaxInt then SourceY :=  MaxInt;
        if SourceY < -MaxInt then SourceY := -MaxInt;
        SourceX := RoundTo(SourceX, -5);
        SourceY := RoundTo(SourceY, -5);

        if (SourceX >= 0.5) and (SourceX <= Source.Bitmap.Width  - 0.5) and
           (SourceY >= 0.5) and (SourceY <= Source.Bitmap.Height - 0.5)
        then begin    // if pixel IS inside original image
          // Assign pixel from original image to the output bitmap
          if SourceX > 1 then SourceX := SourceX - 0.00001;   // to prevent running off
          if SourceY > 1 then SourceY := SourceY - 0.00001;   // the end of the image
          SourceX_Int  := Floor(SourceX - 0.5);
          SourceY_Int  := Floor(SourceY - 0.5);
          SourceLine1  := SourceLines[SourceY_Int];
          SourceLine2  := SourceLines[SourceY_Int + 1];
          DeltaX       := SourceX - SourceX_Int;
          DeltaY       := SourceY - SourceY_Int;
          NewValue     := Interpolate(SourceLine1[SourceX_Int * 3],
                                      SourceLine1[(SourceX_Int + 1) * 3],
                                      SourceLine2[SourceX_Int * 3],
                                      SourceLine2[(SourceX_Int + 1) * 3],
                                      DeltaX, DeltaY);
          NewValue_Int := CapPixelValue(NewValue);
          ResultLine[PixelIndex * 3]       := NewValue_Int;
          NewValue     := Interpolate(SourceLine1[(SourceX_Int * 3) + 1],
                                      SourceLine1[((SourceX_Int + 1) * 3) + 1],
                                      SourceLine2[(SourceX_Int * 3) + 1],
                                      SourceLine2[((SourceX_Int + 1) * 3) + 1],
                                      DeltaX, DeltaY);
          NewValue_Int := CapPixelValue(NewValue);
          ResultLine[(PixelIndex * 3) + 1] := NewValue_Int;
          NewValue     := Interpolate(SourceLine1[(SourceX_Int * 3) + 2],
                                      SourceLine1[((SourceX_Int + 1) * 3) + 2],
                                      SourceLine2[(SourceX_Int * 3) + 2],
                                      SourceLine2[((SourceX_Int + 1) * 3) + 2],
                                      DeltaX, DeltaY);
          NewValue_Int := CapPixelValue(NewValue);
          ResultLine[(PixelIndex * 3) + 2] := NewValue_Int;
        end
        else begin    // if pixel IS NOT inside original image
          ResultLine[PixelIndex * 3]       := Source.GreyPoint;
          ResultLine[(PixelIndex * 3) + 1] := Source.GreyPoint;
          ResultLine[(PixelIndex * 3) + 2] := Source.GreyPoint;
        end           //  if inside original image
      end;            //  for PixelIndex
      if LineIndex mod UpdateInterval = 0
      then begin
        SaveForm.ProgressBar.Position := Ceil(SaveForm.ProgressBar.Max * LineIndex / TempResult.Height);
      end;
    end;              //  for LineIndex

    if TempZoom < 1
    then begin
      X_Offset := Ceil((Result.Width  - (TempResult.Width /  TempZoom))  / 2);
      Y_Offset := Ceil((Result.Height - (TempResult.Height / TempZoom))  / 2);
      Result.Canvas.StretchDraw(Rect(X_Offset, Y_Offset,
                                     X_Offset + Ceil(TempResult.Width /  TempZoom),
                                     Y_Offset + Ceil(TempResult.Height /  TempZoom)),
                                TempResult);
    end
    else begin
      X_Offset := (Result.Width  - TempResult.Width)  div 2;
      Y_Offset := (Result.Height - TempResult.Height) div 2;
      Result.Canvas.Draw(X_Offset, Y_Offset, TempResult);
    end;

  finally
    Screen.Cursor := SaveCursor;  // Restore to normal
    SaveForm.ProgressBar.Position := 0;
    SetLength(SourceLines, 0);
    TempResult.Free;
  end;                //  try
end;


procedure TSaveForm.OKBtnClick(Sender: TObject);
var
  DialogFileNames : String;
begin
  DialogFileNames := '';

  if (FileExists(Image1FullOutputPath) and Image1SourceNameEdit.Enabled)
    then DialogFileNames := DialogFileNames + Image1FullOutputPath + #13 + #13;
  if (FileExists(Image2FullOutputPath) and Image2SourceNameEdit.Enabled)
    then DialogFileNames := DialogFileNames + Image2FullOutputPath + #13 + #13;
  if (FileExists(Image3FullOutputPath) and Image3SourceNameEdit.Enabled)
    then DialogFileNames := DialogFileNames + Image3FullOutputPath + #13;

  if Length(DialogFileNames) > 0
  then begin
    if MessageDlg('The following files will be overwriten ...' +
                   #13 +
                   #13 +
                   DialogFileNames +
                   #13 +
                  'Click OK to proceed or Cancel to return to the Save Dialog' ,
                   mtWarning ,
                   mbOKCancel , 0) = mrCancel then Exit;
  end;

  ModalResult := mrOk;

end;

procedure TSaveForm.JPGQualitySliderChange(Sender: TObject);
begin
  JPGQualityLabel.Caption := IntToStr(JPGQualitySlider.Position) + '%';
end;

procedure TSaveForm.CancelBtnClick(Sender: TObject);
begin
  ModalResult := mrCancel;
end;

procedure TSaveForm.SelectDirBtnClick(Sender: TObject);
var
DialogFileNames : String;
begin
  DialogFileNames := '';
  if Image1SourceNameEdit.Enabled
  then DialogFileNames := DialogFileNames + Image1OutputFileName + '; ';
  if Image2SourceNameEdit.Enabled
  then DialogFileNames := DialogFileNames + Image2OutputFileName + '; ';
  if Image3SourceNameEdit.Enabled
  then DialogFileNames := DialogFileNames + Image3OutputFileName + '; ';
  ChooseDirectoryDialog.FileName := DialogFileNames;
  ChooseDirectoryDialog.Title := 'Choose Output Directory';
  ChooseDirectoryDialog.InitialDir := OutputDirectory;
  if ChooseDirectoryDialog.Execute
  then begin
      OutputDirectory := ExtractFileDir(ChooseDirectoryDialog.FileName);
      OutputDirLabel.Caption := 'Output Directory ...\' + ExtractFileName(OutputDirectory);
      RecalculateBtn.Click;
  end;
end;

procedure ShowFullPath;
var
I       : Integer;
Extract : String;
Result  : String;
begin
  I       := 1;
  Extract := '';
  Result  := '';
  repeat
    while SaveForm.OutputFullDirLabel.Canvas.TextWidth(Extract + 'aa') < SaveForm.OutputFullDirLabel.Width
    do begin
      Extract := Extract + copy(SaveForm.OutputDirectory, I, 1);
      Inc(I);
      if I > Length(SaveForm.OutputDirectory)
      then Break;
    end;
    Result := Result + Extract + ' ';
    Extract := '';
  until I > Length(SaveForm.OutputDirectory);
  SaveForm.OutputFullDirLabel.Caption := Result;
end;

procedure TSaveForm.RecalculateBtnClick(Sender: TObject);
var
DialogFileNames : String;
begin
  DialogFileNames := '';
  OutputFileList.Clear;
  OutputDirEdit.Text := OutputDirectory;
  ShowFullPath;
  OutputDirLabel.Caption := 'Output Directory ...\' + ExtractFileName(OutputDirectory);

  if Image1SourceNameEdit.Enabled
  then begin
    if AppendCheckBox.Checked
    then Image1OutputFileName := Image1SourceNameEdit.Text + AlignedTextEdit.Text
    else Image1OutputFileName := AlignedTextEdit.Text + Image1SourceNameEdit.Text;
    if JPGRadio.Checked
    then Image1OutputFileName := Image1OutputFileName + '.jpg'
    else Image1OutputFileName := Image1OutputFileName + '.bmp';
    OutputFileList.Items.Add(Image1OutputFileName);
    DialogFileNames := DialogFileNames + Image1OutputFileName + '; ';
    Image1FullOutputPath := OutputDirEdit.Text + '\' + Image1OutputFileName;
  end
  else begin
    OutputFileList.Items.Add('');
    Image1FileName := '';
    Image1OutputFileName := '';
    Image1FullOutputPath := ''
  end;

  if Image2SourceNameEdit.Enabled
  then begin
    if AppendCheckBox.Checked
    then Image2OutputFileName := Image2SourceNameEdit.Text + AlignedTextEdit.Text
    else Image2OutputFileName := AlignedTextEdit.Text + Image2SourceNameEdit.Text;
    if JPGRadio.Checked
    then Image2OutputFileName := Image2OutputFileName + '.jpg'
    else Image2OutputFileName := Image2OutputFileName + '.bmp';
    OutputFileList.Items.Add(Image2OutputFileName);
    DialogFileNames := DialogFileNames + Image2OutputFileName + '; ';
    Image2FullOutputPath := OutputDirEdit.Text + '\' + Image2OutputFileName;
  end
  else begin
    OutputFileList.Items.Add('');
    Image2FileName := '';
    Image2OutputFileName := '';
    Image2FullOutputPath := ''
  end;

  if Image3SourceNameEdit.Enabled
  then begin
    if AppendCheckBox.Checked
    then Image3OutputFileName := Image3SourceNameEdit.Text + AlignedTextEdit.Text
    else Image3OutputFileName := AlignedTextEdit.Text + Image3SourceNameEdit.Text;
    if JPGRadio.Checked
    then Image3OutputFileName := Image3OutputFileName + '.jpg'
    else Image3OutputFileName := Image3OutputFileName + '.bmp';
    OutputFileList.Items.Add(Image3OutputFileName);
    DialogFileNames := DialogFileNames + Image3OutputFileName + '; ';
    Image3FullOutputPath := OutputDirEdit.Text + '\' + Image3OutputFileName;
  end
  else begin
    OutputFileList.Items.Add('');
    Image3FileName := '';
    Image3OutputFileName := '';
    Image3FullOutputPath := ''
  end;

  if (FileExists(Image1FullOutputPath)) or
     (FileExists(Image2FullOutputPath)) or
     (FileExists(Image3FullOutputPath))
    then FileOverwriteWarningStaticText.Visible := True
    else FileOverwriteWarningStaticText.Visible := False;

  if (FileExists(Image1FullOutputPath))
  then Image1FileOverwriteMarker.Visible := True
  else Image1FileOverwriteMarker.Visible := False;

  if (FileExists(Image2FullOutputPath))
  then Image2FileOverwriteMarker.Visible := True
  else Image2FileOverwriteMarker.Visible := False;

  if (FileExists(Image3FullOutputPath))
  then Image3FileOverwriteMarker.Visible := True
  else Image3FileOverwriteMarker.Visible := False;

  Form1.SaveDialog1.FileName := DialogFileNames;
  Form1.SaveDialog1.Title := 'Choose Output Directory';
  Form1.SaveDialog1.InitialDir := OutputDirectory;

end;

procedure TSaveForm.OutputFileListClick(Sender: TObject);
begin

  // Clear the Result canvas
  Result.Canvas.Brush.Color := clBlack;
  Result.Canvas.Rectangle(0, 0, Result.Width, Result.Height);

    if OutputFileList.Items.Count > 0 then
      if OutputFileList.Selected[0] and Form1.Image1.HasBitmap and (Image1OutputFileName <> '')
        then begin
          BuildSavePreviewBitmap(Form1.Image1, (Form1.Master = 1), Result);
          SelectedImage := 1;
        end;

    if OutputFileList.Items.Count > 1 then
      if OutputFileList.Selected[1] and Form1.Image2.HasBitmap and (Image2OutputFileName <> '')
        then begin
          BuildSavePreviewBitmap(Form1.Image2, (Form1.Master = 2), Result);
          SelectedImage := 2;
        end;

    if OutputFileList.Items.Count > 2 then
      if OutputFileList.Selected[2] and Form1.Image3.HasBitmap and (Image3OutputFileName <> '')
        then begin
          BuildSavePreviewBitmap(Form1.Image3, (Form1.Master = 3), Result);
          SelectedImage := 3;
        end;

  ImagePreviewPaintBox.Canvas.Draw(0,0,Result);

end;

procedure TSaveForm.FormCreate(Sender: TObject);
begin
  Result := TBitmap.Create;
  Result.PixelFormat        := pf24bit;
  Result.Width :=  ImagePreviewPaintBox.Width;
  Result.Height := ImagePreviewPaintBox.Height;
  Result.Canvas.Brush.Color := clBlack;
  Result.Canvas.Rectangle(0, 0, Result.Width, Result.Height);
  FileOverwriteWarningStaticText.Color := RGB(255,150,150);
  Image1FileOverwriteMarker.Brush.Color := RGB(255,150,150);
  Image2FileOverwriteMarker.Brush.Color := RGB(255,150,150);
  Image3FileOverwriteMarker.Brush.Color := RGB(255,150,150);
  SelectedImage := 0;
end;

procedure TSaveForm.FormDestroy(Sender: TObject);
begin
    Result.Free;
end;

procedure TSaveForm.OutputDirEditChange(Sender: TObject);
begin
  OutputDirEdit.Text := OutputDirectory;
  ShowFullPath;
end;

procedure TSaveForm.ImagePreviewPaintBoxPaint(Sender: TObject);
begin
  ImagePreviewPaintBox.Canvas.Draw(0,0,Result);
end;

procedure TSaveForm.TopBdrEditChange(Sender: TObject);
begin

  if not(BordersCheckBox.Checked)
  then Exit;

  // Clear the Result canvas
  Result.Canvas.Brush.Color := clBlack;
  Result.Canvas.Rectangle(0, 0, Result.Width, Result.Height);

      if (SelectedImage = 1) and BordersCheckBox.Checked
        then begin
          BuildSavePreviewBitmap(Form1.Image1, (Form1.Master = 1), Result);
        end;

      if (SelectedImage = 2) and BordersCheckBox.Checked
        then begin
          BuildSavePreviewBitmap(Form1.Image2, (Form1.Master = 2), Result);
        end;

      if (SelectedImage = 3) and BordersCheckBox.Checked
        then begin
          BuildSavePreviewBitmap(Form1.Image3, (Form1.Master = 3), Result);
        end;

  ImagePreviewPaintBox.Canvas.Draw(0,0,Result);

end;

procedure TSaveForm.BordersCheckBoxClick(Sender: TObject);
begin

  // Clear the Result canvas
  Result.Canvas.Brush.Color := clBlack;
  Result.Canvas.Rectangle(0, 0, Result.Width, Result.Height);

      if (SelectedImage = 1)
        then begin
          BuildSavePreviewBitmap(Form1.Image1, (Form1.Master = 1), Result);
        end;

      if (SelectedImage = 2)
        then begin
          BuildSavePreviewBitmap(Form1.Image2, (Form1.Master = 2), Result);
        end;

      if (SelectedImage = 3)
        then begin
          BuildSavePreviewBitmap(Form1.Image3, (Form1.Master = 3), Result);
        end;

  ImagePreviewPaintBox.Canvas.Draw(0,0,Result);

end;

procedure TSaveForm.Image1SourceNameEditClick(Sender: TObject);
begin
  BuildSavePreviewBitmap(Form1.Image1, (Form1.Master = 1), Result);
  OutputFileList.Selected[0] := True;
  ImagePreviewPaintBox.Canvas.Draw(0,0,Result);
end;

procedure TSaveForm.Image2SourceNameEditClick(Sender: TObject);
begin
  BuildSavePreviewBitmap(Form1.Image2, (Form1.Master = 2), Result);
  OutputFileList.Selected[1] := True;
  ImagePreviewPaintBox.Canvas.Draw(0,0,Result);
end;

procedure TSaveForm.Image3SourceNameEditClick(Sender: TObject);
begin
  BuildSavePreviewBitmap(Form1.Image3, (Form1.Master = 3), Result);
  OutputFileList.Selected[2] := True;
  ImagePreviewPaintBox.Canvas.Draw(0,0,Result);
end;

end.
